	page	240, 132
;KBHAN.ASM	2-Apr-2020
;Replacement interrupt handler for the keyboard. This simply returns an
; integer with bits that indicate which arrow keys are pressed or released.

	public	setkbv, reskbv, getkeys
	.386

;=============================== REAL MODE =====================================

PMODE_TEXT	segment para public use16 'code'
		assume	cs:PMODE_TEXT, ds:dseg

;-------------------------------------------------------------------------------
;This replacement keyboard interrupt handler simply sets or clears bits
; in keydown that indicate whether certain keys are pressed or released.
;
kbhan:	push	ax			;save registers
	push	bx
	push	cx
	push	ds

	mov	ax, dseg		;point to our data
	mov	ds, ax

	in	al, 060h		;read scan code from keyboard
	mov	ch, al			;save it

	in	al, 061h		;send acknowledge signal to keyboard
	or	al, 080h		; by strobing the high bit
	out	061h, al
	and	al, 07fh
	out	061h, al

	mov	al, 20h			;send end-of-interrupt (EOI) to 8259A
	out	20h, al
	sti

	mov	bx, offset kbtbl	;point to scan code table
	mov	cl, nkeys		;set up loop counter

	cmp	byte ptr relflag, 0ffh	;is release flag set? (for AT's only)
	je	short kb50		;jump if so

	cmp	ch, 0f0h		;is AT keyboard sending release code?
	jne	short kb05		;jump if not
	 mov	byte ptr relflag, 0ffh	;set release flag
	 jmp	short kb90		;exit
kb05:	or	ch, ch			;set status
	js	short kb50		;jump if release code

;here if key was pressed
	mov	ax, 00001h		;set bit
kb10:	cmp	ch, [bx]		;test for scan code
	jne	short kb20		;jump if not recognized
	 or	keydown, ax		;set bit corresponding to key
kb20:	rol	ax, 1			;next bit position
	inc	bx			;next scan code
	dec	cl
	jne	short kb10		;loop until all codes tested
	jmp	short kb90		;exit

;here if key was released
kb50:	and	ch, 07fh		;clear MSB to form pressed scan code
	mov	ax, 0fffeh		;set up bit mask
kb60:	cmp	ch, [bx]		;test for scan code
	jne	short kb70		;jump if not recognized
	 and	keydown, ax		;clear bit corresponding to key
kb70:	rol	ax, 1			;next bit position
	inc	bx			;next scan code
	dec	cl
	jne	short kb60		;loop until all codes tested

	mov	byte ptr relflag, 0	;clear release flag

kb90:	pop	ds			;restore registers
	pop	cx
	pop	bx
	pop	ax
	iret


dseg	segment	para public use32 'data'
	align	4

keydown	dw	0			;bit is set if key is down
relflag	db	0			;flag for AT keyboards: key is released

nkeys	equ	13			;number of keys in kbtbl
;table of scan codes			;BIT KEY
kbtbl	db	36h			; 0: right Shift
	db	2Ah			; 1: left Shift
	db	1Dh			; 2: Ctrl
	db	38h			; 3: Alt

	db	46h			; 4: Scroll Lock on
	db	45h			; 5: Num Lock on
	db	3Ah			; 6: Caps Lock on
	db	52h			; 7: Insert

	db	4Dh			; 8: right arrow
	db	50h			; 9: down arrow
	db	4Bh			;10: left arrow
	db	48h			;11: up arrow

	db	01h			;12: escape
dseg	ends

PMODE_TEXT	ends

;============================ PROTECTED MODE ===================================

cseg	segment para public use32 'code'
	assume	cs:cseg, ds:dseg

dseg	segment	para public use32 'data'

;Real mode call structure
		align 4
regs_edi	label	dword
regs_di	dw	0, 0
regs_esi	label	dword
regs_si	dw	0, 0
regs_ebp	label	dword
regs_bp	dw	0, 0
		dd	0
regs_ebx	label	dword
regs_bx	label	word
regs_bl	db	0
regs_bh	db	0, 0, 0
regs_edx	label	dword
regs_dx	label	word
regs_dl	db	0
regs_dh	db	0, 0, 0
regs_ecx	label	dword
regs_cx	label	word
regs_cl	db	0
regs_ch	db	0, 0, 0
regs_eax	label	dword
regs_ax	label	word
regs_al	db	0
regs_ah	db	0, 0, 0
regs_flags	dw 0
regs_es	dw	0
regs_ds	dw	0
regs_fs	dw	0
regs_gs	dw	0
regs_ip	dw	0
regs_cs	dw	0
regs_sp	dw	0
regs_ss	dw	0

kbvec	dw	0, 0			;original keyboard interrupt vector
dseg	ends

;-------------------------------------------------------------------------------
;External routine that returns a value indicating which keys are up or down.
;
getkeys:pop	eax			;pop return address

	movzx	edx, keydown		;push the key states
	push	edx

	push	eax			;restore return address
	ret

;-------------------------------------------------------------------------------
;External routine to set the keyboard interrupt vector (09h) to point to our
; handler.
;
setkbv:	push	edi			;preserve register

	mov	keydown, 0		;assume all keys are released
	mov	byte ptr relflag, 0

	mov	edi, offset regs_edi	;save address of int 09h handler
	mov	word ptr [regs_ax-regs_edi+edi], 3509h
	xor	ecx, ecx		;no parameters on stack
	mov	[regs_ss-regs_edi+edi], cx ;SS:SP = 0, PMODE will provide stack
	mov	[regs_sp-regs_edi+edi], cx
	mov	bx, 21h			;call interrupt 21h
	mov	ax, 300h		;int 31h function 0300h
	int	31h			;do the call to real mode
	mov	bx, [regs_bx-regs_edi+edi]
	mov	word ptr kbvec, bx	;offset
	mov	bx, [regs_es-regs_edi+edi]
	mov	word ptr kbvec+2, bx	;segment

	mov	word ptr [regs_ax-regs_edi+edi], 2509h ;set int 09h vector
	mov	edx, offset kbhan	;load far pointer into break handler
	mov	[regs_dx-regs_edi+edi], dx
	mov	ax, seg kbhan
	mov	[regs_ds-regs_edi+edi], ax
	xor	ecx, ecx		;no parameters on stack
	mov	[regs_ss-regs_edi+edi], cx ;SS:SP = 0, PMODE will provide stack
	mov	[regs_sp-regs_edi+edi], cx
	mov	bx, 21h			;call interrupt 21h
	mov	ax, 300h		;int 31h function 0300h
	int	31h			;do the call to real mode

	pop	edi			;restore register
	ret

;-------------------------------------------------------------------------------
;External routine to restore original keyboard interrupt vector.
;
reskbv:	push	edi			;preserve register

	mov	edi, offset regs_edi	;set original int 09h vector
	mov	word ptr [regs_ax-regs_edi+edi], 2509h
	mov	dx, kbvec		;load far pointer into break handler
	mov	[regs_dx-regs_edi+edi], dx
	mov	ax, kbvec+2		;segment
	mov	[regs_ds-regs_edi+edi], ax
	xor	ecx, ecx		;no parameters on stack
	mov	[regs_ss-regs_edi+edi], cx ;SS:SP = 0, PMODE will provide stack
	mov	[regs_sp-regs_edi+edi], cx
	mov	bx, 21h			;call interrupt 21h
	mov	ax, 300h		;int 31h function 0300h
	int	31h			;do the call to real mode

	pop	edi			;restore register
	ret

cseg	ends
	end
