\plat.xpl	12-Apr-2020
\Side scrolling platform game
\Translated from program in "Eloquent JavaScript" (2nd Ed) by Marijn Haverbeke.
\Sound routines by Larry Fish.

ext	GetKeys, SetKBV, ResKBV;\make arrow keys work like push buttons

int	LevelWidth, LevelHeight,\level dimensions (in grid cells)
	LevelStatus;
def	\LevelStatus\ Null, Lost, Won;
char	Grid(120, 50);		\contains ledges and stationary lava, etc.

\Window into a level scene that fills the screen
int	WindowPos(2);		\position of upper-left corner: X, Y (in pixels)
def	WindowWidth = 640, WindowHeight = 480,	\(pixels)
	MarginWidth = WindowWidth/4,	\bunny's distance from window's edge
	MarginHeight = WindowHeight/4;	\ where window scrolls

def	Cell=16;		\cell size for sprites (square)
def	Scale = float(Cell);	\scale factor

char	Image(WindowWidth*WindowHeight);	\copy of screen image
char	Bkgnd(640*480);		\background image

char	Sprites,		\contains sprite image data
	Palette;		\color palette array for sprites and background

int	PenX, PenY;		\pen position for drawing sprites (pixels)

char	SprLedge(2+16*16),	\sprite images
	SprLedge2(2+16*16),
	SprLava(2+16*16),
	SprLava2(2+16*16),
	SprEgg(2+16*16),
	SprEgg2(2+16*16),
	SprVirus(2+16*16),
	SprEgg3(2+16*16),
	SprBunnyR(2+16*24),	\right facing bunny
	SprBunnyL(2+16*24),	\left facing bunny
	SprBurntR(2+16*24),	\burned rabbit
	SprBurntL(2+16*24);

real	DT;			\delta time; frame display duration (seconds)

real	BunnyPos(2),		\position: X, Y (Grid divisions)
	BunnyDims(2),		\dimensions: width, height (Grid divisions)
	BunnyVel(2);		\velocity: X, Y (Grid divisions/second)
def	BunnyXVel = 7.0,	\velocity for horizontal motion (Grid divs/sec)
	Gravity = 30.0,		\acceleration due to gravity (Grid divs/sec^2)
	JumpVel = -17.0;	\(negative direction is upward)
int	BunnyRight;		\bool: bunny is facing to the right (else left)

def	EggSize = 30,		\max. number of eggs and lava allowed per level
	LavaSize = 30;		\equals sizes of arrays containing these items
int	EggInx, LavaInx;	\indexes into these items' arrays
real	EggPos(EggSize, 2),	\current position: X, Y (Grid divisions)
	EggBasePos(EggSize, 2), \initial position
	EggWobble(EggSize),	\offset for tantalizing up and down motion
	EggDims;		\dimensions: width, height (Grid divisions)
def	WobbleVel = 8.0,  WobbleDist = 0.07; \(Grid divisions)

int	LavaType(LavaSize);	\=, |, v: horizontal, vertical, and dripping
real	LavaPos(LavaSize, 2),	\position: X, Y (Grid divisions)
	LavaVel(LavaSize, 2),	\velocity: X, Y (Grid divisions/second)
	LavaRepeatPos(LavaSize, 2), \initial position for dripping (v) lava
	LavaDims;		\dimensions: width, height (Grid divisions)

def	Pi = 3.14159265358979323846;
string	0;			\use null-terminated strings



proc	Exit(Str);		\Display (error) message and exit program
char	Str;
[ResKBV;			\restore original keyboard handler vector
SetVid($03);			\make sure screen is in normal text mode
Text(0, Str);  CrLf(0);		\display any message
exit;				\return to OS
];	\Exit



func	StrLen(A);		\Return number of characters in an ASCIIZ string
char	A;
int	I;
for I:= 0 to -1>>1 -1 do
	if A(I) = 0 then return I;



func real VectAdd(A, B, C);	\Add two 2D vectors
real	A, B, C;		\A:= B + C
[A(0):= B(0) + C(0);		\VectAdd(A, A, C) => A:= A + C
 A(1):= B(1) + C(1);
return A;
];	\VectAdd



func real VectMul(A, B, S);	\Multiply 2D vector by a scalar
real	A, B, S;		\A:= B * S
[A(0):= B(0) * S;		\VectMul(A, A, S) => A:= A * S
 A(1):= B(1) * S;
return A;
];	\VectMul

\=============================== SOUND ROUTINES ================================

\Play some notes using the Adlib / Sound Blaster's FM chip

def	MaxChan = 9;			\Maximum number of channels
def	SB_BASE = $220,			\Sound Blaster base port address

	DSP_RESET = SB_BASE+$06,	\DSP Reset port
	DSP_READ = SB_BASE+$0A,		\DSP Read data
	DSP_WRITE = SB_BASE+$0C,	\DSP Write data
	DSP_WSTATUS = SB_BASE+$0C,	\DSP Write status
	DSP_RSTATUS = SB_BASE+$0E,	\DSP Read status

	FM_STATUS = SB_BASE+$08,	\FM Status port (read)
	FM_REGISTER = SB_BASE+$08,	\FM Register select port (write)
	FM_DATA = SB_BASE+$09;		\FM Data port (write only)

def	FM_20_A, FM_20_B,		\RegArray index values
	FM_40_A, FM_40_B,
	FM_60_A, FM_60_B,
	FM_80_A, FM_80_B,
	FM_E0_A, FM_E0_B,
	FM_C0, FM_01, FM_08, FM_BD,
	MaxReg;

def	\Define note values
	NoteC, NoteCS, NoteD, NoteDS, NoteE, NoteF,
	NoteFS, NoteG, NoteGS, NoteA, NoteAS, NoteB;

int	Bottom,			\Bottom octave for musical scale
	ModulateOffset,		\Modulator offsets for each channel
	CarrierOffset,		\Carrier offsets for each channel

	ChanNote(MaxChan),	\Array of notes for each channel
	ChanOctave(MaxChan),	\Array of octaves for each channel
	ChanTime(MaxChan),	\Array of times for each channel to be on

	RegArray(MaxReg);	\Array of FM control register values



func	ReadDSP(Port);		\Read from specified DSP Port
int	Port;
int	C;
[for C:= 0 to 1600 do
	if port(DSP_RSTATUS) & $80 then return port(Port);
return 0;
];	\ReadDSP



func	ResetDSP;		\Reset DSP chip (Required for Windows)
int	C;
[port(DSP_RESET):= 1;
for C:= 0 to 256 do [];
port(DSP_RESET):= 0;
return ReadDSP(DSP_READ, 0) = $AA;
];	\ResetDSP



proc	SetFMReg(Reg, Value);	\Set specified register to "Value"
int	Reg, Value;
int	C, T;
[port(FM_REGISTER):= Reg;		\Select register
for C:= 1 to 6 do T:= port(FM_STATUS);	\Delay at least 3.3us
port(FM_DATA):= Value;			\Write data to selected register
for C:= 1 to 35 do T:= port(FM_STATUS);	\Delay at least 23us
];	\SetFMReg



proc	ResetFM;		\Reset all FM registers
int	R;
[for R:= $F5 downto 2 do SetFMReg(R, 0);
SetFMReg($04, $06);
SetFMReg($01, $20);
];	\ResetFM



proc	SetFMDefaults;		\Set sound quality defaults
[RegArray(FM_01):= $20;
RegArray(FM_08):= 0;
RegArray(FM_BD):= 0;

RegArray(FM_20_A):= $01;   RegArray(FM_20_B):= $01;
RegArray(FM_40_A):= $10;   RegArray(FM_40_B):= $00;
RegArray(FM_60_A):= $F0;   RegArray(FM_60_B):= $F0;
RegArray(FM_80_A):= $77;   RegArray(FM_80_B):= $77;
RegArray(FM_E0_A):= 0;     RegArray(FM_E0_B):= 0;
RegArray(FM_C0):= 0;
];	\SetFMDefaults



proc	SetTimbre(Chan, Array);	\Set sound quality for a channel
int	Chan, Array;
int	MO, CO;
[MO:= ModulateOffset(Chan);
 CO:= CarrierOffset(Chan);

SetFMReg($01, Array(FM_01));
SetFMReg($08, Array(FM_08));
SetFMReg($BD, Array(FM_BD));

SetFMReg($20+MO, Array(FM_20_A));
SetFMReg($20+CO, Array(FM_20_B));

SetFMReg($40+MO, Array(FM_40_A));
SetFMReg($40+CO, Array(FM_40_B));

SetFMReg($60+MO, Array(FM_60_A));
SetFMReg($60+CO, Array(FM_60_B));

SetFMReg($80+MO, Array(FM_80_A));
SetFMReg($80+CO, Array(FM_80_B));

SetFMReg($E0+MO, Array(FM_E0_A));
SetFMReg($E0+CO, Array(FM_E0_B));

SetFMReg($C0, Array(FM_C0));
SetFMReg($40+CO, Array(FM_40_B)+5);
];	\SetTimbre



func	SetChannel(Chan, Note, Octave, Array);	\Set sound channel
int	Chan, Note, Octave, Array;
int	F1, F2;
[SetTimbre(Chan, Array);
F1:= Bottom(Note);
F2:= Swap(F1) & $03;   F2:= F2 ! Octave<<2;   F2:= F2 ! $20;
SetFMReg($A0+Chan, F1);
SetFMReg($B0+Chan, F2);
return F1;
];	\SetChannel



proc	PlayChord;		\Play chord in array
int	Chan, F;
[for Chan:= 0 to MaxChan-1 do
    if ChanNote(Chan) # -1 then
	F:= SetChannel(Chan, ChanNote(Chan), ChanOctave(Chan), RegArray);
];	\PlayChord



proc	SetReg(RegValues);	\Set timbre register array values
\ in this order:
\ 20A, 20B,  40A, 40B,  60A, 60B,  80A, 80B,  E0A, E0B,  C0, 01, 08, BD
int	RegValues;
int	I;
for I:= 0 to MaxReg-1 do RegArray(I):= RegValues(I);



proc	GotEgg;			\Make "got egg" sound
int	Reg;
[Reg:= [$05, $01, $4E, $80, $DA, $F6, $15, $13, $00, $00, $0A, $20, $00, $00];
SetReg(Reg);
ChanNote(0):= NoteC;   ChanOctave(0):= 5;   ChanTime(0):= 2;   PlayChord;
ChanNote(0):= NoteE;   ChanOctave(0):= 5;   ChanTime(0):= 2;   PlayChord;
ChanNote(0):= NoteG;   ChanOctave(0):= 5;   ChanTime(0):= 2;   PlayChord;
];	\GotEgg



proc	Loser;			\Make "loser" sound
int	Reg;
[Reg:= [$31, $21, $94, $00, $33, $A2, $73, $97, $00, $00, $0E, $20, $00, $00];
SetReg(Reg);

\G7 Chord
ChanNote(0):= NoteG;   ChanOctave(0):= 5;   ChanTime(0):= 10;
ChanNote(1):= NoteB;   ChanOctave(1):= 5;
ChanNote(2):= NoteD;   ChanOctave(2):= 5;
ChanNote(3):= NoteF;   ChanOctave(3):= 5;
PlayChord;

\C Chord
ChanNote(0):= NoteC;   ChanOctave(0):= 5;   ChanTime(0):= 10;
ChanNote(1):= NoteE;   ChanOctave(1):= 5;
ChanNote(2):= NoteG;   ChanOctave(2):= 5;
PlayChord;
];	\Loser



proc	PlaySound(Snd);		\Play specified sound
int	Snd;
int	I;
[Bottom:= [$156, $16B, $181, $198, $1B0, $1CA, $1E5,
	   $202, $220, $241, $263, $287, $2A3];
ModulateOffset:= [$00, $01, $02, $08, $09, $0A, $10, $11, $12];
CarrierOffset:=  [$03, $04, $05, $0B, $0C, $0D, $13, $14, $15];
for I:= 0 to MaxChan-1 do ChanNote(I):= -1;
SetFMDefaults;		\terminate any previous sound to enable new sound
ResetFM;

case Snd of
  1:	GotEgg
other	Loser;
];	\PlaySound

\===============================================================================

proc	FadeIn;	\Slowly increase intensities of palette colors to final values
int	I, J;
char	Pal(256*3);
[for I:= 1 to 64 do		\64 steps take about one second
	[for J:= 0 to 256*3-1 do Pal(J):= Palette(J)*I/64;
	WaitForVSync;
	for J:= 0 to 256-1 do SetPalette(J, Pal(J*3+0), Pal(J*3+1), Pal(J*3+2));
	];
];	\FadeIn



proc	FadeOut; \Slowly decrease intensities of palette colors to black
int	I, J;
char	Pal(256*3);
[for I:= 63 downto 0 do		\64 steps take about one second
	[for J:= 0 to 256*3-1 do Pal(J):= Palette(J)*I/64;
	WaitForVSync;
	for J:= 0 to 256-1 do SetPalette(J, Pal(J*3+0), Pal(J*3+1), Pal(J*3+2));
	];
];	\FadeOut



proc	CycleLava;		\Cycle lava palette colors
int	Reg, RGB, TRGB;
[TRGB:= GetPalette($F0);
for Reg:= $F0 to $F7 do
	[RGB:= GetPalette(Reg+1);
	SetPalette(Reg, RGB>>16, RGB>>8, RGB);
	];
SetPalette(Reg, TRGB>>16, TRGB>>8, TRGB);
];	\CycleLava

\-------------------------------------------------------------------------------

proc	Move(X, Y);		\Reposition drawing pen
int	X, Y;
[PenX:= X;  PenY:= Y];



proc	DrawSprite(X0, Y0, Spr); \Draw a sprite to Image array
int	X0, Y0;	\coordinates of upper-left corner (pixels)
char	Spr;	\address of sprite data
int	X, Y, W, H;
[W:= Spr(0);			\get width and height (in pixels)
 H:= Spr(1);
Spr:= Spr+2;
for Y:= Y0 to Y0+H-1 do
    for X:= X0 to X0+W-1 do
	[if Spr(0) then		\black (0) is transparent; don't plot it
	  if Y>=0 & Y<WindowHeight then
	    if X>=0 & X<WindowWidth then
	      Image(X + Y*WindowWidth):= Spr(0);
	Spr:= Spr+1;
	];

PenX:= PenX + Cell;		\move to next character position
];	\DrawSprite



proc	ShowLevel;		\Display current level thru a full-screen window
int	X, Y, I, Color, PX, PY;
[CopyMem(Image, Bkgnd, 640*480);

\Convert pixels to char cells
for Y:= WindowPos(1)/Cell to (WindowPos(1)+WindowHeight-1)/Cell do
	[Move(-(rem(WindowPos(0)/Cell)),
		(Y-WindowPos(1)/Cell)*Cell-(rem(WindowPos(1)/Cell)));
	for X:= WindowPos(0)/Cell to (WindowPos(0)+WindowWidth-1)/Cell do
		\Convert Plan symbols to sprites
		case Grid(X,Y) of
		   0:	[PenX:= PenX + Cell];
		  ^x:	DrawSprite(PenX, PenY, SprLedge);
		  ^z:	DrawSprite(PenX, PenY, SprLedge2);
		  ^!:	DrawSprite(PenX, PenY, SprLava)
		other	[IntOut(6, X);  Text(6, ", ");
			IntOut(6, Y);  Text(6, ", ");
			HexOut(6, Grid(X, Y)); if ChIn(1) then; Exit("")];
	];

PX:= fix(BunnyPos(0)*Scale) - WindowPos(0);
PY:= fix(BunnyPos(1)*Scale) - WindowPos(1);
Move(PX, PY);
if LevelStatus = Lost then
	DrawSprite(PenX, PenY, if BunnyRight then SprBurntR else SprBurntL)
else	DrawSprite(PenX, PenY, if BunnyRight then SprBunnyR else SprBunnyL);

\Show eggs
for I:= 0 to EggInx-1 do
	[PX:= fix(EggPos(I,0)*Scale) - WindowPos(0);
	 PY:= fix(EggPos(I,1)*Scale) - WindowPos(1);
	Move(PX, PY);
	case rem(I/3) of
	  1: Color:= SprEgg;
	  2: Color:= SprEgg2
	other Color:= SprEgg3;
	DrawSprite(PenX, PenY, Color);
	];

\Show lava
for I:= 0 to LavaInx-1 do
	[PX:= fix(LavaPos(I,0)*Scale) - WindowPos(0);
	 PY:= fix(LavaPos(I,1)*Scale) - WindowPos(1);
	Move(PX, PY);
	DrawSprite(PenX, PenY, if LavaType(I)= ^= then SprVirus else SprLava2);
	];

Paint(0, 0, WindowWidth, WindowHeight, Image, WindowWidth);
];	\ShowLevel



proc	MoveWindow;		\Move Window to track Bunny's position
int	P;
[P:= fix(BunnyPos(0)*Scale);	\convert Bunny's position to pixels
if P < WindowPos(0)+MarginWidth then
	[WindowPos(0):= P - MarginWidth;
	if WindowPos(0) < 0 then WindowPos(0):= 0;
	];
if P >= WindowPos(0)+WindowWidth-MarginWidth then
	[WindowPos(0):= P - (WindowWidth-MarginWidth);
	P:= LevelWidth*Cell - WindowWidth;
	if WindowPos(0) >= P then WindowPos(0):= P;
	];
P:= fix(BunnyPos(1)*Scale);
if P < WindowPos(1)+MarginHeight then
	[WindowPos(1):= P - MarginHeight;
	if WindowPos(1) < 0 then WindowPos(1):= 0;
	];
if P >= WindowPos(1)+WindowHeight-MarginHeight then
	[WindowPos(1):= P - (WindowHeight-MarginHeight);
	P:= LevelHeight*Cell - WindowHeight;
	if WindowPos(1) >= P then WindowPos(1):= P;
	];
];	\MoveWindow

\-------------------------------------------------------------------------------

proc	BunnyTouched;		\Handle bunny touching moving lava and eggs
int	I;
[for I:= 0 to LavaInx-1 do					\  +---+
	if BunnyPos(0)+BunnyDims(0) >= LavaPos(I,0) &		\  |   |
	   BunnyPos(0) <= LavaPos(I,0)+LavaDims(0) &   		\  |  ++--+
	   BunnyPos(1)+BunnyDims(1) >= LavaPos(I,1) & 		\  +--++  |
	   BunnyPos(1) <= LavaPos(I,1)+LavaDims(1) then		\     |   |
	   	[LevelStatus:= Lost;				\     +---+
	   	PlaySound(0);
	   	return];
for I:= 0 to EggInx-1 do
	if BunnyPos(0)+BunnyDims(0) >= EggPos(I,0) &
	   BunnyPos(0) <= EggPos(I,0)+EggDims(0) &
	   BunnyPos(1)+BunnyDims(1) >= EggPos(I,1) &
	   BunnyPos(1) <= EggPos(I,1)+EggDims(1) then
		[EggPos(I,0):= -100.0;		\remove egg
		PlaySound(1);
		];
for I:= 0 to EggInx-1 do			\if any eggs left then return
	if EggPos(I,0) >= 0.0 then return;
LevelStatus:= Won;				\else level has been won
];	\BunnyTouched



func	ObstacleAt(Pos, Dims);	\Return what bunny or lava ran into
real	Pos, Dims;	\position, dimensions, 2D vectors
int	StartX, EndX, StartY, EndY;
int	X, Y;
[StartX:= fix(Floor(Pos(0)));		\	xxxxxxxx	level is bounded
EndX:= fix(Ceil(Pos(0) + Dims(0)));	\	x      x	by ledges & lava
StartY:= fix(Floor(Pos(1)));		\	x      x
EndY:= fix(Ceil(Pos(1) + Dims(1)));	\	!!!!!!!!
if StartX<0 ! EndX>LevelWidth-1 ! StartY<0 then return ^x;	\ledge
if EndY > LevelHeight then return ^!;				\lava
for Y:= StartY to EndY-1 do		\check if bunny hit fixed lava in Grid
    for X:= StartX to EndX-1 do		\ or moving lava hit ledge or fixed lava
	if Grid(X, Y) then return Grid(X, Y);
return 0;							\none
];	\ObstacleAt



proc	MoveBunnyX(Keys);	\Move bunny horizontally
int	Keys;		\bit array for Shift, Ctrl, and Alt keys
real	NewPos(2);
int	Obstacle;
[BunnyVel(0):= 0.0;		\assume no motion
if Keys & $0400 then		\left arrow
	[BunnyVel(0):= BunnyVel(0) - BunnyXVel;  BunnyRight:= false];
if Keys & $0100 then		\right arrow
	[BunnyVel(0):= BunnyVel(0) + BunnyXVel;  BunnyRight:= true];
NewPos(0):= BunnyPos(0) + BunnyVel(0)*DT;	\x:= x + v*dt
NewPos(1):= BunnyPos(1);			\tentative new position

Obstacle:= ObstacleAt(NewPos, BunnyDims);
if Obstacle then	\you're toast if you ran into some lava
	[if Obstacle = ^! then [LevelStatus:= Lost;  PlaySound(0)]]
else	BunnyPos(0):= NewPos(0);		\move to empty position
];	\MoveBunnyX



proc	MoveBunnyY(Keys);	\Move bunny vertically
int	Keys;		\bit array for Shift, Ctrl, and Alt keys
real	NewPos(2);
int	Obstacle;
[BunnyVel(1):= BunnyVel(1) + Gravity*DT;	\v = v + g*dt
NewPos(1):= BunnyPos(1) + BunnyVel(1)*DT;	\y = y + v*dt
NewPos(0):= BunnyPos(0);			\tentative new position

Obstacle:= ObstacleAt(NewPos, BunnyDims);
if Obstacle then
	[if Obstacle = ^! then [LevelStatus:= Lost;  PlaySound(0)];
	if (Keys&$0800) # 0 & BunnyVel(1) > 0.0 then
		BunnyVel(1):= JumpVel		\jump off ledge
	else	BunnyVel(1):= 0.0;		\stand on ledge
	]
else	BunnyPos(1):= NewPos(1);		\move to empty position
];	\MoveBunnyY



proc	MoveBunny;		\Move bunny using keyboard keys
int	Keys;		\bit array for Shift, Ctrl, and Alt keys
[Keys:= GetKeys;

MoveBunnyX(Keys);
MoveBunnyY(Keys);
BunnyTouched;		\handle bunny colliding with moving lava or eggs
if LevelStatus = Lost then
	BunnyPos(1):= BunnyPos(1) + DT;	\bunny drops dead!
if GetKeys & $1000 then Exit("");		\Esc aborts program
MoveWindow;
];	\MoveBunny



proc	MoveEggs;		\Make eggs alluring by wobbling them
int	I;
[for I:= 0 to EggInx-1 do
	[EggWobble(I):= EggWobble(I) + WobbleVel*DT;
	EggPos(I,1):= EggBasePos(I,1) + WobbleDist * Sin(EggWobble(I));
	];
];	\MoveEggs



proc	MoveLava;		\Move horizontal, vertical, and dripping lava
real	NewPos(2);
int	I;
[for I:= 0 to LavaInx-1 do
	[VectMul(NewPos, LavaVel(I), DT);		\dx = v * dt
	 VectAdd(NewPos, NewPos, LavaPos(I));		\x = x + dx
	if ObstacleAt(NewPos, LavaDims) = 0 then	\move to empty position
		[LavaPos(I,0):= NewPos(0);  LavaPos(I,1):= NewPos(1)]
	else if LavaRepeatPos(I,0) # 0.0 then		\dripping lava
		[LavaPos(I,0):= LavaRepeatPos(I,0);	\only "v" lava repeats
		 LavaPos(I,1):= LavaRepeatPos(I,1)]
	else	VectMul(LavaVel(I), LavaVel(I), -1.0);	\reverse lava direction
	];
];	\MoveLava



proc	RunLevel;		\Play a level of the game
real	FinishDelay;	\delay at end of level, for death scene, etc.
int	T0, T1;
real	LT;		\lava cycle timer
[FinishDelay:= 1.0;	\delay one second
ShowLevel;
FadeIn;			\turn the lights up slowly
T0:= GetTime;
LT:= 0.0;
repeat	T1:= GetTime;
	DT:= float(T1-T0)/1e6;
	if DT > 0.1 then DT:= 0.1;	\prevents falling thru ledges if slow
	T0:= T1;
	if LT >= 0.3 then
		[LT:= 0.0;  CycleLava]
	else	LT:= LT + DT;
	ShowLevel;
	MoveBunny;
	MoveEggs;
	MoveLava;
	if LevelStatus # Null then	\won or lost: run timer
		FinishDelay:= FinishDelay - DT;
until	FinishDelay <= 0.0;
FadeOut;
Clear;			\erase screen, ready for next level
];	\RunLevel

\-------------------------------------------------------------------------------

proc	MakeBunny(X, Y);	\Initialize bunny parameters for current level
int	X, Y;
[BunnyPos(0):= float(X);  BunnyPos(1):= float(Y) - 0.5;
BunnyDims(0):= 0.8;  BunnyDims(1):= 1.5;
BunnyVel(0):= 0.0; BunnyVel(1):= 0.0;
];	\MakeBunny



proc	MakeEgg(X, Y);		\Initialize egg parameters for current level
int	X, Y;
int	I;
[EggDims:= [0.9, 0.9];
I:= EggInx;
EggPos(I,0):= float(X);  EggPos(I,1):= float(Y);
\\VectAdd(EggPos(I), EggPos(I), [0.2, 0.1]);
EggBasePos(I,0):= EggPos(I,0);  EggBasePos(I,1):= EggPos(I,1);
EggWobble(I):= float(Ran(1000)) / 1000.0 * 2.0 * Pi;
if I < EggSize-1 then EggInx:= EggInx+1
else	Exit("Too many eggs.");
];	\MakeEgg



proc	MakeLava(X, Y, Type);	\Initialize moving lava parameters for level
int	X, Y, Type;
int	I;
[I:= LavaInx;
LavaType(I):= Type;
LavaPos(I,0):= float(X);  LavaPos(I,1):= float(Y);
LavaDims:= [1.0, 1.0];
LavaRepeatPos(I,0):= 0.0;  LavaRepeatPos(I,1):= 0.0;
case Type of
  ^=:	[LavaVel(I,0):= 2.0;  LavaVel(I,1):= 0.0];	\horizontally moving
  ^|:	[LavaVel(I,0):= 0.0;  LavaVel(I,1):= 2.0];	\vertically moving
  ^v:	[LavaVel(I,0):= 0.0;  LavaVel(I,1):= 3.0;	\dripping
	LavaRepeatPos(I,0):= LavaPos(I,0);  LavaRepeatPos(I,1):= LavaPos(I,1);
	]
other	[];
if I < LavaSize-1 then LavaInx:= LavaInx+1
else	Exit("Too much lava.");;
];	\MakeLava



func	MakeLevel(Plan);	\Initialize parameters for current Plan level
char	Plan;
int	X, Y, Ch, FieldType;
[LevelWidth:= StrLen(@Plan(1,0));
LevelHeight:= @Plan(0,0);
WindowPos(0):= 0;  WindowPos(1):= 0;  
EggInx:= 0;
LavaInx:= 0;
LevelStatus:= Null;
BunnyRight:= true;

for Y:= 0 to LevelHeight-1 do
    [for X:= 0 to LevelWidth-1 do
	[FieldType:= 0;
	Ch:= Plan(Y+1, X);	\(note reversed argument order)
	case Ch of
	  ^@:	MakeBunny(X, Y);
	  ^o:	MakeEgg(X, Y);
	  ^=, ^|, ^v:		\types of moving lava
		MakeLava(X, Y, Ch);
	  ^x, ^!, ^z:		\ledges and stationary lava
		FieldType:= Ch;
	  ^ :	[]		\space characters convert to zeros
	other	Exit("Unrecognized Plan character.");
	Grid(X, Y):= FieldType;
	];
    ];
MoveWindow;			\make sure bunny is in the display window
];	\MakeLevel

\-------------------------------------------------------------------------------

proc	GetSprite(X0, Y0, W, H, Spr);	\Load sprite array from Sprites array
int	X0, Y0,	\coordinates in Sprites array of upper-left corner of sprite
	W, H;	\width and height of sprite (pixels)
char	Spr;	\address of array in which to load sprite data: Spr(2+W*H)
int	X, Y;
[Spr(0):= W;
 Spr(1):= H;
Spr:= Spr+2;
for Y:= Y0 to Y0+H-1 do
    for X:= X0 to X0+W-1 do
	[Spr(0):= Sprites(X+Y*6*16);	\Sprites dimensions are (6x2)*16 pixels
	Spr:= Spr+1;
	];
];	\GetSprite



proc	LoadSprites;		\Load sprite arrays from Sprites data array
int	I, J;
[include platspr;		\constant arrays containing Sprites data
GetSprite(0*16, 0, Cell, Cell, SprEgg);
GetSprite(1*16, 0, Cell, Cell, SprLedge2);
GetSprite(2*16, 0, Cell, Cell, SprLava);
GetSprite(3*16, 0, Cell, Cell, SprVirus);

GetSprite(0*16, 16, Cell, Cell, SprLava2);
GetSprite(1*16, 16, Cell, Cell, SprLedge);
GetSprite(2*16, 16, Cell, Cell, SprEgg2);
GetSprite(3*16, 16, Cell, Cell, SprEgg3);

GetSprite(4*16, 0, 16, 24, SprBunnyR);	\right facing
GetSprite(5*16, 0, 16, 24, SprBurntR);

SprBunnyL(0):= SprBunnyR(0);		\make left-facing sprite
SprBunnyL(1):= SprBunnyR(1);
SprBurntL(0):= SprBurntR(0);		\make left-facing sprite
SprBurntL(1):=  SprBurntR(1);
for J:= 0 to 23 do
    for I:= 0 to 15 do
	[SprBunnyL(2+I+J*16):= SprBunnyR(2+15-I+J*16);
	 SprBurntL(2+I+J*16):= SprBurntR(2+15-I+J*16);
	];
];	\LoadSprites



func	LoadBmp(FileName);	\Load specified .bmp file into Bkgnd array
char	FileName;		\Returns 'true' if successful
\A compatible graphic mode is assumed to have been set up with SetVid.
\Reference: Microsoft MSDN Library; and "Graphics File Formats" by Kay & Levine
int	FD,		\file descriptor (handle)
	Offset,		\offset from start of file to image data (bytes)
	HeadSize,	\header size: 40=Windows format; 12=OS/2 (old) format
	Width, Height,	\file image dimensions (pixels)
	Depth,		\file image depth (bits)
	R, G, B,	\red, green, blue (0..255)
	X, Y,		\screen coordinates (pixels)
	I, J, T,	\indexes and temporary scratch
	CTblSize;	\number of bytes in BMP's color table

[Trap(false);			\turn off error trapping (if error, don't abort)
FD:= FOpen(FileName, 0);	\get input file descriptor
if GetErr \#0\ then return false;
Trap(true);
FSet(FD, ^I);			\set device 3 to opened file
OpenI(3);

if ChIn(3) # ^B then return false;	   \file must begin with "BM" (Bit Map)
if ChIn(3) # ^M then return false;
for I:= 2 to 9 do X:= ChIn(3);		   \ 2: skip unused header info
Offset:= ChIn(3) + ChIn(3)<<8;		   \10: offset from start to image data
T:= ChIn(3);  T:= ChIn(3);		   \12: skip unused header info
HeadSize:= ChIn(3) + ChIn(3)<<8;	   \14: size of header (40 or 12)
T:= ChIn(3);  T:= ChIn(3);		   \16: skip high bytes of header size
Width:= ChIn(3) + ChIn(3)<<8;		   \18: get image Width (in pixels)
if HeadSize = 12 then			   \OS/2 (old) format:
	[Height:= ChIn(3) + ChIn(3)<<8;	   \20: get image Height (in pixels)
	T:= ChIn(3);  T:= ChIn(3);	   \22: skip "number of image planes"
	Depth:= ChIn(3) + ChIn(3)<<8;	   \24: bits per pixel
	]
else					   \Windows (normal) format:
	[T:= ChIn(3);  T:= ChIn(3);	   \20: skip unused high bytes of Width
	Height:= ChIn(3) + ChIn(3)<<8;	   \22: get image Height (in pixels)
	T:= ChIn(3);  T:= ChIn(3);	   \24: skip unused high bytes of Height
	T:= ChIn(3);  T:= ChIn(3);	   \26: skip "number of image planes"
	Depth:= ChIn(3) + ChIn(3)<<8;	   \28: bits per pixel
	for I:= 30 to 53 do T:= ChIn(3);   \30: skip rest of header
	];

\Read in and set palette registers with color table if it's provided
CTblSize:= Offset - 14 - HeadSize;	\number of bytes of color table data
if CTblSize > 256*4 then CTblSize:= 256*4;
I:= 0;  J:= 0;
while J < CTblSize do
	[B:= ChIn(3);
	 G:= ChIn(3);
	 R:= ChIn(3);
	 T:= ChIn(3);
	 J:= J+4;
	];

\Note: scan lines are padded to end on a 4-byte boundary
for Y:= Height-1 downto 0 do		\bmps are upside down (!)
	[case Depth of
	  8:	[for X:= 0 to Width-1 do
			Bkgnd(X + Y*640):= ChIn(3);
		]
	other	return false;		\unsupported depth
	];
FClose(FD);
return true;				\no errors detected
];	\LoadBmp



int	N, Level, GameLevels;
[include levels.xpl;
SetVid($101);
for N:= 0 to 256-1 do		\turn off the lights
	SetPalette(N, 0, 0, 0);
TrapC(true);			\prevent Ctrl+C from aborting
SetKBV;				\replace keyboard interrupt handler vector

LoadBmp("bkgnd.bmp");
LoadSprites;			\also sets up Palette array

ResetDSP;
ResetFM;

for Level:= IntIn(8) to 4-1 do
	repeat	MakeLevel(GameLevels(Level));
		RunLevel;
		DelayUS(1_000_000);
		ResetFM;	\silence lingering sounds
	until	LevelStatus = Won;
Exit("            ^M^J             
         ^M^J            ");
]
